!+
! Module: emeventsproj_mod
!
! Authors:  Jean Ballet - SAp   CEA Saclay
!          
! Description: Project EPIC-MOS events back onto an image
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emeventsproj_mod

  use edusoft_mod
  use eventAttributes
  use cal
  use datafilter  
  use dal
  use param  
  use errorHandling

  implicit none 
  
  private 

! Names of the task parameters (2nd string for the keywords)

    character(len=*), parameter :: inevent     = "eventset", &
                                   outimage    = "evimageset", &
                                   reject      = "rejectbadevents", &
                                   energy      = "projectenergy"

! Name of the extensions

    character(len=*), parameter :: event_ext="EVENTS"

! Maximum size of the read-out area

    integer,parameter :: dx1=5, dx2=5, dy1=0, dy2=2
    integer,parameter :: x_size_window=EMOS_MAX_X+dx1+dx2
    integer,parameter :: y_size_window=EMOS_MAX_Y+dy1+dy2

! Routine name

    character(len=*), parameter :: taskname = "emeventsproj"

  public emeventsproj

contains 
  
  !---------------------------------------------------------------------
  subroutine emeventsproj
  !---------------------------------------------------------------------

    type(sas_evt), dimension(:), allocatable :: evt

! Declarations for the DAL

    type(TableT) :: ev_tab
    type(ArrayT) :: ev_map

! Function flags
! rejectbad  : Reject the events with rejection flag set when projecting ?
! projenergy : Project the energy value (if no, project the counts) ?
! fromodf    : Is this an events file straight from the ODF ?

    logical :: rejectbad, projenergy, fromodf
    integer :: nevents

! Read parameters

    call getparamvalues(ev_tab, ev_map, rejectbad, projenergy, fromodf)

! Select the appropriate events

    call selectEvents(ev_tab, rejectbad, fromodf)

! Read the data

    nevents = numberOfRows(ev_tab)
    allocate(evt(nevents))
    if ( nevents > 0 )  call readevents(ev_tab, evt)
    call release(parent(ev_tab))

! Build the map

    call buildmap(evt, ev_map, projenergy, fromodf)

! Release everything

    deallocate(evt)
    call release(parent(ev_map))

  end subroutine emeventsproj

! Build the map by backprojecting all events

  subroutine buildmap(evt, ev_map, projenergy, fromodf)

! evt    : Array of event structure
! ev_map : Handle to the output array
! projenergy : Project the energy value (if no, project the counts) ?
! fromodf    : Is this an events file straight from the ODF ?

    type(sas_evt), dimension(:),intent(in) :: evt
    type(ArrayT), intent(in)               :: ev_map
    logical, intent(in)                    :: projenergy, fromodf

! Interface with the DAL/CAL

    integer(kind=int32), dimension(:,:), pointer   :: map
    integer(kind=int8),  dimension(:,:,:), pointer :: patterns
    integer(kind=int16)                            :: eduThreshold

! Local variables

    integer(kind=int32), dimension(-2:x_size_window,-2:y_size_window) :: image
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1) :: patabove
    integer :: npattern, jr

! Read the pattern library

    call CAL_getEventPatterns(patterns, eduThreshold)

! Transpose the pattern library
! Set the 2s and the central 1 to 0

    do npattern = 0,edu_npat-1
       do jr = -1,1
          patabove(:,jr,npattern) = patterns(npattern,jr+2,1:3)
       enddo
    enddo
    where (patabove > 1)
       patabove = 0
    endwhere
    patabove(0,0,:) = 0

! Initialise the image to 0

    image = 0

! Project the events on a larger image to avoid unpleasant surprises

    if (projenergy) then
       call projectEventsEnergy(evt, patabove, image)
    else
       call projectEventsCounts(evt, patabove, image)
    endif
    write(errstr,*) "Total signal projected to map: ", sum(image)
    call message(AppMsg,Verbose)

! Write the relevant part of the image to the output file
! This is offset because of (-2,-2) in projectEvents

    map => int32Array2Data(ev_map)
    if (fromodf) then
       map = image(dx1:EMOS_MAX_X+dx1-1,dy1:EMOS_MAX_Y+dy1-1)
    else
       map = image(-1:EMOS_MAX_X-2,-1:EMOS_MAX_Y-2)
    endif
    write(errstr,*) "Total signal within CCD area : ", sum(map)
    call message(AppMsg,Verbose)

  end subroutine buildmap

! Select the events appropriate to be projected

  subroutine selectEvents(ev_tab, rejectbad, fromodf)

! ev_tab : Handle to the input table
! rejectbad  : Reject the events with rejection flag set when projecting ?
! fromodf    : Is this an events file straight from the ODF ?

    type(TableT), intent(inout) :: ev_tab
    logical, intent(in)         :: rejectbad, fromodf

    character(len=80) :: expression, reject
    integer(kind=int32) :: testqual
    integer :: nrow1

    write(errstr,*) "Number of events in input file: ", numberOfRows(ev_tab)
    call message(AppMsg,Verbose)

! Make the expression for initFilter
! The selection on position is different before emevents
! because RAWX,RAWY are offset by dx1+1, dy1+1 there

    if (fromodf) then
       write(expression,100)  dx1+1, EMOS_MAX_X+dx1+2, dy1+1, EMOS_MAX_Y+dy1+2
    else
       write(expression,100)  0, EMOS_MAX_X+1, 0, EMOS_MAX_Y+1
    endif
 100   format("(RAWX-",I3,")*(",I3,"-RAWX)>0 && (RAWY-",I3,")*(",I3,"-RAWY)>0")
    if (rejectbad) then

! Do not reject pixels outside the field of view or truncated events

       testqual = EVATT_EMOS_REJECTION_MASK
       if (iand(testqual,EVATT_OUT_OF_FOV) > 0)  &
          testqual = testqual - EVATT_OUT_OF_FOV
       if (iand(testqual,EVATT_REJECTED_BY_GATTI) > 0)  &
          testqual = testqual - EVATT_REJECTED_BY_GATTI
       write(reject,'("(FLAG & 0x",Z8.8,")==0")') testqual
       expression = trim(expression) // " && " // reject
    endif
    write(errstr,*) "Call initFilter with expression ", expression
    call message(AppMsg,Verbose)
    call initFilter( expression )       

! Filter ev_tab

    nrow1 = doFilter( ev_tab )
    write(errstr,*) "Number of good events in input file: ", nrow1
    call message(AppMsg,Sparse)

  end subroutine selectEvents

! Write the events data to a structure

  subroutine readevents(ev_tab, evt)

! ev_tab : Handle to the input table
! evt    : Array of event structure

    type(TableT), intent(in) :: ev_tab
    type(sas_evt), dimension(:),intent(out) :: evt

! Name of the columns

    character(len=8), parameter :: rawx = "RAWX    ", rawy = "RAWY    ", &
                                   ene1 = "ENERGYE1", ene2 = "ENERGYE2", &
                                   patt = "PATTERN "

    real(kind=double), dimension(:), allocatable :: timebuf

    allocate(timebuf(numberOfRows(ev_tab)))
    timebuf = 0.d0

! Only RAWX, RAWY, PATTERN, ENERGYE1 and ENERGYE2 are useful

    call copy_block(ev_tab, rawx, timebuf)
    evt%rawx = int(timebuf+sign(0.5D0,timebuf),int16)
    call copy_block(ev_tab, rawy, timebuf)
    evt%rawy = int(timebuf+sign(0.5D0,timebuf),int16)
    call copy_block(ev_tab, patt, timebuf)
    evt%pattern = int(timebuf+sign(0.5D0,timebuf),int8)
    call copy_block(ev_tab, ene1, timebuf)
    evt%e1 = int(timebuf+sign(0.5D0,timebuf),int16)
    call copy_block(ev_tab, ene2, timebuf)
    evt%e2 = int(timebuf+sign(0.5D0,timebuf),int16)

    deallocate(timebuf)

  end subroutine readevents

! Copy column into array with possible type change

  subroutine copy_block(ev_tab, colname, buffer)

! ev_tab  : handle to the input table for the DAL
! colname : column name
! buffer  : output array (double)

    type(TableT),  intent(in)  :: ev_tab
    character(len=*), intent(in) :: colname
    real(kind=double), dimension(:), intent(out) :: buffer

    type(ColumnT) :: in_col
    integer :: coltype

! Possible types of columns

    real(kind=double),   dimension(:), pointer   :: dbledata
    real(kind=single),   dimension(:), pointer   :: sngldata
    integer(kind=int32), dimension(:), pointer   :: longdata
    integer(kind=int16), dimension(:), pointer   :: intgdata
    integer(kind=int8),  dimension(:), pointer   :: bytedata

! Read input column

    in_col  = column(ev_tab, colname, Read)
    coltype = columnDataType(in_col)
    select case(coltype)
    case (Real64)
       dbledata => real64Data(in_col)
       buffer = dbledata
    case (Real32)
       sngldata => real32Data(in_col)
       buffer = sngldata
    case (Integer32)
       longdata => int32Data(in_col)
       buffer = longdata
    case (Integer16)
       intgdata => int16Data(in_col)
       buffer = intgdata
    case (Integer8)
       bytedata => int8Data(in_col)
       buffer = bytedata
    case default
       write(errstr,*) "input column ",qualifiedName(in_col),  &
                       " has an unknown type ", coltype
       call error("copy_data 01")
    end select
    call release(in_col)

  end subroutine copy_block

! Parameter reading

  subroutine getparamvalues(ev_tab, ev_map, rejectbad, projenergy, fromodf)

! ev_tab : Handle to the input event table
! ev_map : Handle to the output image
! rejectbad  : Reject the events with rejection flag set when projecting ?
! projenergy : Project the energy value (if no, project the counts) ?
! fromodf    : Is this an events file straight from the ODF ?

    type(TableT), intent(out) :: ev_tab
    type(ArrayT), intent(out) :: ev_map
    logical, intent(out)      :: rejectbad, projenergy, fromodf

    type(DataSetT) :: ev_set, im_set

    character(len=1024) :: image_out, event_in
    character(len=80)   :: comm, inst

! Open input event file
! Should check instrument and mode

    event_in = stringParameter(inevent)
    write(errstr,*) "Opening input event file : ",trim(event_in)
    call message(AppMsg,Sparse)

! Cloning the file for selectlib

    ev_set = clone(event_in, 'emeventsproj.temp', Temp)
    fromodf = .not. hasBlock(ev_set, event_ext)

    if (fromodf) then
       ev_tab = table(ev_set, 0)  ! Extension name depends on mode
    else
       ev_tab = table(ev_set, event_ext)
    endif
    inst = stringAttribute(ev_tab,"INSTRUME")
    if (inst(1:4) .ne. 'EMOS') then
       write(errstr,*) "The program works only for XMM EPIC MOS, not for ",  &
                       trim(inst)
       call error("getparamvalues 01")
    endif
    if (stringAttribute(ev_tab,"DATATYPE") .ne. "IMAGE.EL") then
       write(errstr,*) "The program works only for Imaging mode, not for ",  &
                       stringAttribute(ev_tab,"DATATYPE"), " data"
       call error("getparamvalues 02")
    endif
    call CAL_setState(block(ev_tab))

! Switches

    projenergy = booleanParameter(energy)
    if (projenergy) then
       call message(AppMsg,Sparse,"The events' energy will be projected.")
       comm = "Cumulative energy"
    else
       call message(AppMsg,Sparse,"The counts will be projected.")
       comm = "2-D histogram"
    endif

    rejectbad  = booleanParameter(reject)
    if (rejectbad .and. fromodf) then
       write(errstr,*) "Events straight from the ODF are not flagged yet. ",  &
                       "Cannot reject flagged events"
       call warning("getparamvalues 10")
       rejectbad = .false.
    endif
    if (rejectbad) then
       if (fromodf) then
          write(errstr,*) "Events straight from the ODF are not flagged yet.",&
                          " Cannot reject flagged events"
          call warning("getparamvalues 10")
          rejectbad = .false.
          comm = comm // " of all events"
       else
          call message(AppMsg,Sparse, &
                       "Events flagged for rejection will not be projected.")
          comm = comm // " of good events"
       endif
    else
       call message(AppMsg,Sparse, &
                    "Events flagged for rejection will be projected.")
       comm = comm // " of all events"
    endif

! Open output image file

    image_out = stringParameter(outimage)
    write(errstr,*) "Opening output image file : ",trim(image_out)
    call message(AppMsg,Sparse)

    im_set = dataSet(image_out, Create)
    call setAttributes(attributable(im_set),attributable(ev_set))
    ev_map = addArray(im_set, "PRIMARY", Integer32,  &
             (/EMOS_MAX_X,EMOS_MAX_Y/), comm)
    call setAttributes(attributable(ev_map), attributable(ev_tab))
    call deleteAttribute(ev_map,"EXTVER")

  end subroutine getparamvalues

end module emeventsproj_mod
